
/*
  RegiTime displays last write date and time of registry keys and their subkeys.
  RegiTime can also display keys with last write data and time in given data-time range.

  Usage: RegiTime <*|HKCC|HKCR|HKCU|HKLM|HKUS|Path> [YYYY/MM/DD-HH:MM:SS YYYY/MM/DD-HH:MM:SS] [Machine]
  Examples:
   1. Display date-time for HKEY_LOCAL_MACHINE and its subkeys on \\server:
      regitime HKLM \\server 
   2. Display date-time for HKEY_CLASSES_ROOT\*\shellex and its subkeys:
      regitime HKCR\*\shellex
   3. Display date-time of all keys:
      regitime * >bigfile.txt
   4. Display date-time for HKEY_LOCAL_MACHINE and its subkeys last written in 2003 on \\server:
      regitime HKLM 2003 2004 \\server
   5. Display date-time for HKEY_CLASSES_ROOT\*\shellex and its subkeys last written in March 2004:
      regitime HKEY_CLASSES_ROOT\*\shellex 2004-3 2003/4
   6. Display date-time of all keys last written on April 18, 2004:
      regitime * 2004/4/18 2004-04-19
*/


#include <stdlib.h>
#include <stdio.h>
#include <windows.h>
#include <tchar.h>


#define KEYNAMESIZE 32768
TCHAR g_szRegKeyName[KEYNAMESIZE];
DWORD g_szRegKeyNameSize;
LPCTSTR g_szhKeyPredef[] = {_T("HKCR"), _T("HKCU"), _T("HKLM"), _T("HKUS"), _T("HKPD"), _T("HKCC"), NULL};
BOOL g_bTimeRange = FALSE;
FILETIME g_StartTime, g_EndTime;


BOOL __stdcall RegiTimeToFileTime(LPCTSTR lpRegiTime, FILETIME *lpFileTime) {
  LPTSTR p;
  SYSTEMTIME st = {0, 1, 0, 1, 0, 0, 0, 0};

  st.wYear = _tcstoul(lpRegiTime, &p, 10);
  if(*p) {
    st.wMonth = _tcstoul(++p, &p, 10);
    if(*p) {
      st.wDay = _tcstoul(++p, &p, 10);
      if(*p) {
        st.wHour = _tcstoul(++p, &p, 10);
        if(*p) {
          st.wMinute = _tcstoul(++p, &p, 10);
          if(*p) {
            st.wSecond = _tcstoul(++p, &p, 10);
          }
        }
      }
    }
  }
  return(SystemTimeToFileTime(&st, lpFileTime));
}

BOOL __stdcall FileTimeToRegiTime(FILETIME *lpFileTime, LPTSTR lpRegiTime) {
  BOOL Result;
  SYSTEMTIME st;

  Result = FileTimeToSystemTime(lpFileTime, &st);
  if(Result) {
    Result = _stprintf(lpRegiTime, _T("%04u/%02u/%02u-%02u:%02u:%02u"), st.wYear, st.wMonth, st.wDay, st.wHour, st.wMinute, st.wSecond);
  }
  return(Result);
}

VOID __stdcall PrintKeyTime(FILETIME *lpFileTime) {
  BOOL bHigherOrAboveStart = TRUE, bBelowEnd = TRUE;
  TCHAR szRegiTime[32];

  FileTimeToLocalFileTime(lpFileTime, lpFileTime);
  if(FileTimeToRegiTime(lpFileTime, szRegiTime)) {
    if(g_bTimeRange) {
      if(lpFileTime->dwHighDateTime < g_StartTime.dwHighDateTime) {
        bHigherOrAboveStart = FALSE;
      }
      else {
        if(lpFileTime->dwHighDateTime == g_StartTime.dwHighDateTime) {
          bHigherOrAboveStart = lpFileTime->dwLowDateTime >= g_StartTime.dwLowDateTime;
        }
      }
      if(lpFileTime->dwHighDateTime > g_EndTime.dwHighDateTime) {
        bBelowEnd = FALSE;
      }
      else {
        if(lpFileTime->dwHighDateTime == g_EndTime.dwHighDateTime) {
          bBelowEnd = lpFileTime->dwLowDateTime < g_EndTime.dwLowDateTime;
        }
      }
    }
    if(bHigherOrAboveStart && bBelowEnd) {
      _tprintf(_T("%s %s\n"), szRegiTime, g_szRegKeyName);
    }
  }
}

VOID __stdcall EnumKeyTime(HKEY hKeyParent, LPCTSTR lpKeyChild) {
  DWORD i, n, l_SKPos;
  LONG r;
  FILETIME ft;

  if(RegOpenKeyEx(hKeyParent, lpKeyChild, 0, KEY_ENUMERATE_SUB_KEYS, &hKeyParent) == ERROR_SUCCESS) {
    g_szRegKeyNameSize--;
    l_SKPos = KEYNAMESIZE -g_szRegKeyNameSize;
    g_szRegKeyName[l_SKPos -1] = '\\';
    for(n = g_szRegKeyNameSize, i = 0; (r = RegEnumKeyEx(hKeyParent, i, g_szRegKeyName +l_SKPos, &n, NULL, NULL, NULL, &ft)) != ERROR_NO_MORE_ITEMS; n = g_szRegKeyNameSize = KEYNAMESIZE -l_SKPos, i++) {
      if(r == ERROR_SUCCESS) {
        PrintKeyTime(&ft);
        g_szRegKeyNameSize -= n;
        EnumKeyTime(hKeyParent, g_szRegKeyName +l_SKPos);
      }
    }
    RegCloseKey(hKeyParent);
  }
}

BOOL __stdcall RegiTime(LPCTSTR lpMachineName, HKEY hKeyPredef, LPCTSTR lpSubKey) {
  BOOL Result = FALSE;
  HKEY hKeyPredefCopy, hKeyRoot;
  FILETIME ft;

  if(RegConnectRegistry(lpMachineName, hKeyPredef, &hKeyPredefCopy) == ERROR_SUCCESS) {
    g_szRegKeyNameSize = KEYNAMESIZE -_stprintf(g_szRegKeyName, lpSubKey ? _T("%s\\%s") : _T("%s%s"), g_szhKeyPredef[(DWORD)hKeyPredef-(DWORD)HKEY_CLASSES_ROOT], lpSubKey ? lpSubKey : _T(""));
    if(RegOpenKeyEx(hKeyPredefCopy, lpSubKey, 0, KEY_QUERY_VALUE, &hKeyRoot) == ERROR_SUCCESS) {
      if(RegQueryInfoKey(hKeyRoot, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, &ft) == ERROR_SUCCESS) {
        Result = TRUE;
        PrintKeyTime(&ft);
      }
      RegCloseKey(hKeyRoot);
      EnumKeyTime(hKeyPredefCopy, lpSubKey);
    }
    RegCloseKey(hKeyPredefCopy);
  }
  return(Result);
}

int _tmain(int argc, TCHAR **argv) {
  LPCTSTR lpMachineName;
  DWORD i;

  if(argc < 4) {
    if(argc <= 1) {
      return(0);
    }
    lpMachineName = argc > 2 ? argv[2] : NULL;
  }
  else {
    #define ErrorTimeString _T("ERROR: Cannot convert time '%s'\n")
    if(!RegiTimeToFileTime(argv[2], &g_StartTime)) {
      return(_tprintf(ErrorTimeString, argv[2]));
    }   
    if(!RegiTimeToFileTime(argv[3], &g_EndTime)) {
      return(_tprintf(ErrorTimeString, argv[3]));
    }
    lpMachineName = argc > 4 ? argv[4] : NULL;
    g_bTimeRange = TRUE;
  }


  if(argv[1][0] == '*') {
    RegiTime(lpMachineName, HKEY_CLASSES_ROOT, NULL);
    RegiTime(lpMachineName, HKEY_CURRENT_USER, NULL);
    RegiTime(lpMachineName, HKEY_LOCAL_MACHINE, NULL);
    RegiTime(lpMachineName, HKEY_USERS, NULL);
    RegiTime(lpMachineName, HKEY_CURRENT_CONFIG, NULL);
  }
  else {
    for(i = 0; g_szhKeyPredef[i]; i++) {
      if(_tcsnicmp(g_szhKeyPredef[i], argv[1], 4) == 0) {
        break;
      }
    }
    if(g_szhKeyPredef[i]) {
      RegiTime(lpMachineName, (HKEY)(i+(DWORD)HKEY_CLASSES_ROOT), argv[1][4] && argv[1][5] ? &argv[1][5] : NULL);
    }
    else {
      return(_tprintf(_T("ERROR: Unsupported HKEY '%s'\n"), argv[1]));
    }
  }

  return(0);
}
